/*
 *  OpenGL 3.x Demonstration
 *
 *  for the "Computergraphics 2" lab
 *
 *  This program demonstrates the basic OpenGL 3.x pipeline and usage.
 *  It requires OpenGL 3.x context-able graphics-hardware and drivers
 *  (except programm modified as noted).
 *  The Demo is only a DEMONSTRATION and an actual application should NOT
 *  be implemented this way (efficiency, maintenance)!
 *
 *  Institute of Computer Graphics and Algorithms
 *  Vienna University of Technology
 *  Vienna, Austria
 *
 *  March 2010
 *
 *  ginquo, vogi, wimmer
 */


// system includes
#include <iostream>
#include <fstream>

// local includes
#include "Program.hpp"
#include "GL/glew.h"
#include "GL/glfw.h"

// GLM includes
//#include "GLM/glm.hpp" -> don't need it, it's already included in the header-file
#include "GLM/gtx.hpp"
#include "GLM/gtc.hpp"
#include "GLM/gtx/type_ptr.hpp"


// global variables
const int winWidth  = 800;
const int winHeight = 600;

double last_time = 0.0;
glm::mat4 perspective, view, model;

// shader handles
GLuint vertex_shader_handle = 0;
GLuint fragment_shader_handle = 0;
GLint shader_program = 0;

// define our light
glm::vec3 light_position(0.0f, 8.0f, 3.0f);
glm::vec4 light_color(1.0f, 1.0f, 1.0f, 1.0f);
glm::vec4 ambient_color(0.1f, 0.1f, 0.1f, 1.0f);




/*
 *  checkGLError
 *
 *  Basic OpenGL Error-Handling function.
 *  If encountered an error it prints an error-message to the console.
 *
 */
void checkGLError()
{
	GLenum error = glGetError();
	if (error != GL_NO_ERROR)
	{
		switch (error)
		{
			case GL_INVALID_ENUM:
				std::cout << "GL: enum argument out of range." << std::endl;
				break;
			case GL_INVALID_VALUE:
				std::cout << "GL: Numeric argument out of range." << std::endl;
				break;
			case GL_INVALID_OPERATION:
				std::cout << "GL: Operation illegal in current state." << std::endl;
				break;
			case GL_INVALID_FRAMEBUFFER_OPERATION:
				std::cout << "GL: Framebuffer object is not complete." << std::endl;
				break;
			case GL_OUT_OF_MEMORY:
				std::cout << "GL: Not enough memory left to execute command." << std::endl;
				break;
			default:
				std::cout << "GL: Unknown error." << std::endl;
		}
	}
}


/*
 *  init
 *
 *  Does all the initializing stuff: create a window with OpenGL-context,
 *                                   init GLEW,
 *                                   (test OpenGL,)
 *                                    load and compile shaders
 */
void init()
{
	// --- init GLFW and set up our window ---
	if (!glfwInit())
	{
		std::cout << "Could not init GLFW!" << std::endl;
		exit(1);
	} else {
		std::cout << "GLFW initialized" << std::endl;
	}

	// Note: The following lines of code are just that long because we test for the lastest
	//       supported OpenGL-version on your system. You do not need to do that.

    // Set flags before window creation
    glfwOpenWindowHint(GLFW_WINDOW_NO_RESIZE, GL_TRUE); // make window size fixed

	// set flags for getting newest possible OpenGL-context
    glfwOpenWindowHint(GLFW_OPENGL_VERSION_MAJOR, 3);
    glfwOpenWindowHint(GLFW_OPENGL_VERSION_MINOR, 2);
    glfwOpenWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	// test if context available, if not, try next lowest OpenGL-version...
    if (glfwOpenWindow(winWidth, winHeight,  // window size
                       0, 0, 0, 0,           // set color-bits RGBA automatic
                       24, 8,                // depth and stencil
                       GLFW_WINDOW))         // windowed mode
	{
		std::cout << "OpenGL 3.2 core context created" << std::endl;
    } else
	{
        glfwOpenWindowHint(GLFW_OPENGL_VERSION_MAJOR, 3);
        glfwOpenWindowHint(GLFW_OPENGL_VERSION_MINOR, 0);

		glfwOpenWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_COMPAT_PROFILE);
        glfwOpenWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

		if (glfwOpenWindow(winWidth, winHeight,  // window size
						   0, 0, 0, 0,           // set color-bits RGBA automatic
						   24, 8,                // depth and stencil
						   GLFW_WINDOW))         // windowed mode
		{
			std::cout << "OpenGL 3.0 forward compatible context created" << std::endl;
        } else
		{
            glfwOpenWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_FALSE);

			if (glfwOpenWindow(winWidth, winHeight,  // window size
							   0, 0, 0, 0,           // set color-bits RGBA automatic
							   24, 8,                // depth and stencil
							   GLFW_WINDOW))         // windowed mode
			{
                std::cout << "OpenGL backward compatible context created" << std::endl;
            } else
			{
                std::cout << "Failed to create OpenGL window!" << std::endl;

                glfwTerminate();
                exit(1);
            }
        }
    }
	glfwSetWindowTitle("OpenGL 3.x Demo");
	// ---


	// --- init GLEW ---
	// Note: first, an OpenGL-context has to be created before initializing GLEW!
	if (glewInit() != GLEW_OK)
	{
		std::cout << "Could not init GLEW!" << std::endl;
		exit(1);
	} else {
		std::cout << "GLEW initialized" << std::endl;
	}


	// --- let's test if OpenGL works as expected ---
	std::cout << "\nTesting:" << std::endl;

    int context_flags, profile;

	// retrieve our OpenGL-version
	std::cout << "OpenGL Version: " << glGetString(GL_VERSION) << std::endl;

    // check if we have a core-profile
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &profile);
	if (profile == GL_CONTEXT_CORE_PROFILE_BIT) {
        std::cout << "Having a core-profile" << std::endl;
    } else {
        std::cout << "Having a compatibility-profile" << std::endl;
    }

    // check if context is forward compatible
    glGetIntegerv(GL_CONTEXT_FLAGS, &context_flags);
    if (context_flags & GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT) {
        std::cout << "Context is forward compatible." << std::endl;
    } else {
        std::cout << "Context is NOT forward compatible." << std::endl;
    }

    // try doing a fixed-function operation. This should create an error in our context.
    std::cout << "Doing illegal operation (glTranslate) now..." << std::endl;
    glTranslatef(1.0f, 0.0f, 0.0f);

    if (glGetError() != GL_NO_ERROR)
        std::cout << "Illegal operation caused an error message. :)\n" << std::endl;
    else
        std::cout << "Illegal operation did NOT cause an error message. :(\n" << std::endl;

	// --- done testing out OpenGL ---


	// --- Do some setup for OpenGL ---

	// enable back-face culling
    glEnable(GL_CULL_FACE);

	// enable depth-buffer
    glEnable(GL_DEPTH_TEST);

	// enable filling of front an back-faces if desired
	//glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// set our viewport
	glViewport(0, 0, winWidth, winHeight);


	// --- load and compile shader ---

	// -- vertex shader --
	std::ifstream vs_file("shader/simple_shader.vert", std::ifstream::in);
	std::string vertex_shader_source = std::string(std::istreambuf_iterator<char>(vs_file), std::istreambuf_iterator<char>());

	// create handle
	vertex_shader_handle = glCreateShader(GL_VERTEX_SHADER);

	// convert to const c-string and attach to shader-handle
	const GLchar *vs_c_str = vertex_shader_source.c_str();
	glShaderSource(vertex_shader_handle, 1, &vs_c_str, NULL);

	// compile shader
	glCompileShader(vertex_shader_handle);

	// do error-checking
	GLint succeeded;
	glGetShaderiv(vertex_shader_handle, GL_COMPILE_STATUS, &succeeded);

	if (!succeeded || !glIsShader(vertex_shader_handle) )
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetShaderiv(vertex_shader_handle, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetShaderInfoLog(vertex_shader_handle, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "Vertex-shader compile error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "Vertex-shader loaded and compiled successfully." << std::endl;
	}

	// -- fragment shader --
	std::ifstream fs_file("shader/simple_shader.frag", std::ifstream::in);
	std::string fragment_shader_source = std::string(std::istreambuf_iterator<char>(fs_file), std::istreambuf_iterator<char>());

	// create handle
	fragment_shader_handle = glCreateShader(GL_FRAGMENT_SHADER);

	// convert to const c-string and attach to shader-handle
	const GLchar *fs_c_str = fragment_shader_source.c_str();
	glShaderSource(fragment_shader_handle, 1, &fs_c_str, NULL);

	// compile shader
	glCompileShader(fragment_shader_handle);

	// do error-checking
	glGetShaderiv(fragment_shader_handle, GL_COMPILE_STATUS, &succeeded);

	if (!succeeded || !glIsShader(fragment_shader_handle) )
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetShaderiv(fragment_shader_handle, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetShaderInfoLog(fragment_shader_handle, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "Fragment-shader compile error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "Fragment-shader loaded and compiled successfully." << std::endl;
	}

	// -- link shader-program --
	shader_program = glCreateProgram();

    // attach shaders
	glAttachShader(shader_program, vertex_shader_handle);
	glAttachShader(shader_program, fragment_shader_handle);

	// before linking, define output of fragment-shader
	glBindFragDataLocation(shader_program, 0, "fragColor");

    // link
    glLinkProgram(shader_program);

    // check for errors
    glGetProgramiv(shader_program, GL_LINK_STATUS, &succeeded);

    if (!succeeded || !glIsProgram(shader_program))
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetProgramiv(shader_program, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetProgramInfoLog(shader_program, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "Shader-program link error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "Shader-program linked successfully.\n" << std::endl;
	}


	// --- init matrices ---

	// calculate the aspect ratio of the window and create a perspective matrix
	// with a 60 degree field of view
    float aspect = float(winWidth)/float(winHeight);
    perspective = glm::perspective(60.0f, aspect, 0.1f, 100.0f);

    // move camera 4 units backward, i.e. movie the scene 4 units forward
	view = glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, -4));

    // initialize the model matrix to identity
	model = glm::mat4(1.0f);


	// set last frame time
	last_time = glfwGetTime();


	// --- done with initializing stuff ---
}


/*
 *  update
 *
 *  Called each frame before rendering the scene.
 *  Updates the scene corresponding to user-input and time passed.
 *
 */
void update()
{
    // calculate time difference from last frame
    double time_now  = glfwGetTime();
    double time_diff = time_now - last_time;
    last_time = time_now;

    // update the model-matrix to do a continuous rotation of the model
	model = glm::rotate(model, float(20.0 * time_diff), glm::vec3(0, 1, 0));
	model = glm::rotate(model, float(20.0 * time_diff), glm::vec3(1, 0, 0));
}


/*
 *  render
 *
 *  Called each frame and renders the scene.
 *
 */
void render()
{
    // Clear buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Bind shader
    glUseProgram(shader_program);

    // set light uniforms
    GLint light_position_uniform = glGetUniformLocation(shader_program, "light_position");
    GLint light_color_uniform    = glGetUniformLocation(shader_program, "light_color");
    GLint ambient_color_uniform  = glGetUniformLocation(shader_program, "ambient_color");

	glUniform3fv(light_position_uniform, 1, glm::gtx::type_ptr::value_ptr(light_position));
    glUniform4fv(light_color_uniform,    1, glm::gtx::type_ptr::value_ptr(light_color));
    glUniform4fv(ambient_color_uniform,  1, glm::gtx::type_ptr::value_ptr(ambient_color));

    // set matrix-uniforms
    GLint perspective_uniform = glGetUniformLocation(shader_program, "perspective");
    GLint view_uniform        = glGetUniformLocation(shader_program, "view");
    GLint model_uniform       = glGetUniformLocation(shader_program, "model");

	glUniformMatrix4fv(perspective_uniform, 1, GL_FALSE, glm::gtx::type_ptr::value_ptr(perspective));
	glUniformMatrix4fv(view_uniform,        1, GL_FALSE, glm::gtx::type_ptr::value_ptr(view));
	glUniformMatrix4fv(model_uniform,       1, GL_FALSE, glm::gtx::type_ptr::value_ptr(model));

    // set attributes
    GLint vertex_attrib = glGetAttribLocation(shader_program, "vertex");
    GLint normal_attrib = glGetAttribLocation(shader_program, "normal");
    GLint color_attrib  = glGetAttribLocation(shader_program, "color");

	glEnableVertexAttribArray(vertex_attrib);
    glEnableVertexAttribArray(normal_attrib);
    glEnableVertexAttribArray(color_attrib);

    glVertexAttribPointer(vertex_attrib, 4, GL_FLOAT, GL_FALSE, 0, vertices);
    glVertexAttribPointer(normal_attrib, 3, GL_FLOAT, GL_FALSE, 0, normals);
    glVertexAttribPointer(color_attrib,  4, GL_FLOAT, GL_FALSE, 0, colors);

    // draw the geometry
	glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_INT, indices);

    // disable attribute-pointers
    glDisableVertexAttribArray(vertex_attrib);
    glDisableVertexAttribArray(normal_attrib);
    glDisableVertexAttribArray(color_attrib);

    // unbind shader-program
    glUseProgram(0);

	// done rendering frame, swap buffers
    glfwSwapBuffers();
}


/*
 *  cleanup
 *
 *  Called before exiting the application.
 *  Cleanup (i.e. free resources) before the application ends.
 *
 */
void cleanup()
{
    // delete shader-objects and shader-program
    glDeleteShader(vertex_shader_handle);
    glDeleteShader(fragment_shader_handle);
    glDeleteProgram(shader_program);

	// tell glfw that we are finished
	glfwTerminate();
}


/*
 *  main
 *
 *  The main-function. Program-control is done here.
 *  Calls other methods: init(), update(), render(), cleanup()
 *
 */
void main()
{
	std::cout << "OpenGL 3.x Demo\nInstitute of Computer Graphics and Algorithms\nVienna University of Technology\nVienna, Austria\n" << std::endl;

	init();

	bool running = true;
    while (running)
	{
		update();
		render();

		checkGLError();
		running = !glfwGetKey(GLFW_KEY_ESC) && glfwGetWindowParam(GLFW_OPENED);
    }

	cleanup();
	exit(0);
}
